import os
import re
import polars as pl
import polars.selectors as cs
from pathlib import Path
from tqdm.notebook import tqdm
from .. import config as cfg
from ..utils import has_gpu
from ..data_processing import feature_aggregation as fa
from ..database.queries import (
    experiment_metadata_sql_query,
    plate_layout_sql_query,
)
from ..logger import (
    log_info,
    log_warning,
)


def get_cell_morphology_ref(
    name: str,
    filter: dict = None,
):
    query = experiment_metadata_sql_query(
        name,
        cfg.DATABASE_SCHEMA,
        cfg.CELL_MORPHOLOGY_METADATA_TYPE,
    )
    df = pl.read_database(query, cfg.DB_URI)

    if filter is None:
        return df
    conditions = []
    # Iterate over each key-value pair in the filter dictionary
    for key, values in filter.items():
        # Create an OR condition for each value associated with a key
        key_conditions = [pl.col(key).str.contains(val) for val in values]
        combined_key_condition = key_conditions[0]
        for condition in key_conditions[1:]:
            combined_key_condition = combined_key_condition | condition
        conditions.append(combined_key_condition)
    # Combine all conditions with AND
    final_condition = conditions[0]
    for condition in conditions[1:]:
        final_condition = final_condition & condition
    # Apply the condition to the DataFrame
    return df.filter(final_condition)


def get_cell_morphology_data(
    cell_morphology_ref_df: pl.DataFrame,
    aggregation_level="cell",
    aggregation_method=None,
    path_to_save: str = "data",
    use_gpu=False,
):
    if aggregation_method is None:
        aggregation_method = {
            "cell": "median",
            "site": "median",
            "well": "median",
            "plate": "mean",
            "compound": "mean",
        }
    object_file_names = cfg.OBJECT_FILE_NAMES
    plate_acq_id = cfg.DATABASE_SCHEMA["EXPERIMENT_PLATE_ACQID_COLUMN"]
    plate_acq_name = cfg.DATABASE_SCHEMA["EXPERIMENT_PLATE_AQNAME_COLUMN"]

    # Create output directory if it doesn't exist
    saving_dir = Path(path_to_save)
    saving_dir.mkdir(parents=True, exist_ok=True)

    # Set up progress bar for feedback
    total_iterations = cell_morphology_ref_df.height * len(object_file_names)
    progress_bar = tqdm(total=total_iterations, desc="Processing")

    all_dataframes = []

    for index, plate_metadata in enumerate(
        cell_morphology_ref_df.iter_rows(named=True)
    ):
        # Print separator and progress info
        separator = "\n" if index else ""
        log_info(
            (
                f"{separator}{'_'*50}"
                f"\nProcessing plate {plate_metadata[plate_acq_name]} ({index + 1} of {cell_morphology_ref_df.height}):"
            )
        )

        # Define and check for existing output files
        output_filename = f"{saving_dir}/{plate_metadata[plate_acq_id]}_{plate_metadata[plate_acq_name]}.parquet"
        if os.path.exists(output_filename):
            log_info(f"File already exists, reading data from: {output_filename}")
            existing_df = pl.read_parquet(output_filename)
            all_dataframes.append(existing_df)
            progress_bar.update(len(object_file_names))
            continue

        # Load and process feature datasets
        object_feature_dataframes = {}
        unusful_col_pattern = r"^(FileName|PathName|ImageNumber|Number_Object_Number)"

        for object_file_name in object_file_names:
            object_feature_file_path = f"{plate_metadata[cfg.DATABASE_SCHEMA['EXPERIMENT_RESULT_DIRECTORY_COLUMN']]}{object_file_name}.parquet"

            # Read the parquet file and adjust column names
            columns_names = pl.scan_parquet(object_feature_file_path).columns
            object_feature_df = pl.read_parquet(
                object_feature_file_path,
                columns=[
                    col
                    for col in columns_names
                    if not re.match(unusful_col_pattern, col)
                ],
            ).rename({"ObjectNumber": "id"})
            object_name = object_file_name.split("_")[-1]
            object_feature_df.columns = [
                f"{col}_{object_name}" for col in object_feature_df.columns
            ]

            object_feature_dataframes[object_name] = object_feature_df
            log_info(
                f"\tReading features {object_feature_df.shape} - {object_name}: \t{object_feature_file_path}"
            )

            progress_bar.update(1)

        log_info("Merging the data")
        # Join nuclei and cell data on specified columns
        df_combined = object_feature_dataframes["cells"].join(
            object_feature_dataframes["nuclei"],
            left_on=[
                "Metadata_AcqID_cells",
                "Metadata_Barcode_cells",
                "Metadata_Well_cells",
                "Metadata_Site_cells",
                "id_cells",
            ],
            right_on=[
                "Metadata_AcqID_nuclei",
                "Metadata_Barcode_nuclei",
                "Metadata_Well_nuclei",
                "Metadata_Site_nuclei",
                "Parent_cells_nuclei",
            ],
            how="left",
            suffix="_nuclei",
        )

        # Further join with cytoplasm data on specified columns
        df_combined = df_combined.join(
            object_feature_dataframes["cytoplasm"],
            left_on=[
                "Metadata_AcqID_cells",
                "Metadata_Barcode_cells",
                "Metadata_Well_cells",
                "Metadata_Site_cells",
                "id_cells",
            ],
            right_on=[
                "Metadata_AcqID_cytoplasm",
                "Metadata_Barcode_cytoplasm",
                "Metadata_Well_cytoplasm",
                "Metadata_Site_cytoplasm",
                "Parent_cells_cytoplasm",
            ],
            how="left",
            suffix="_cytoplasm",
        )

        # Renaming columns for better consistency
        rename_map = {
            "Metadata_AcqID_cells": "Metadata_AcqID",
            "Metadata_Barcode_cells": "Metadata_Barcode",
            "Metadata_Well_cells": "Metadata_Well",
            "Metadata_Site_cells": "Metadata_Site",
            "Children_cytoplasm_Count_cells": "Cell_cytoplasm_count",
            "Children_nuclei_Count_cells": "Cell_nuclei_count",
        }
        df_combined = df_combined.rename(rename_map)

        # Create ImageID column by concatenating other columns
        image_id = (
            df_combined["Metadata_AcqID"]
            + "_"
            + df_combined["Metadata_Barcode"]
            + "_"
            + df_combined["Metadata_Well"]
            + "_"
            + df_combined["Metadata_Site"]
        ).alias("ImageID")
        df_combined = df_combined.with_columns([image_id])

        # Create CellID column by concatenating other columns
        cell_id = (df_combined["ImageID"] + "_" + df_combined["id_cells"]).alias(
            "CellID"
        )
        df_combined = df_combined.with_columns([cell_id])

        drop_map = [
            "Children_cytoplasm_Count_nuclei",
            "Parent_precells_cells",
            "Parent_nuclei_cytoplasm",
            "id_cells",
            "id_nuclei",
            "id_cytoplasm",
        ]
        df_combined = df_combined.drop(drop_map)

        # Ensure data type consistency for certain columns
        cast_cols = [
            pl.col("Metadata_AcqID").cast(pl.Utf8),
            pl.col("Metadata_Site").cast(pl.Utf8),
        ]
        df_combined = df_combined.with_columns(cast_cols)

        # ordering the columns

        morphology_feature_cols = df_combined.select(
            cs.by_dtype(pl.NUMERIC_DTYPES)
        ).columns
        morphology_feature_cols.remove("Cell_nuclei_count")
        morphology_feature_cols.remove("Cell_cytoplasm_count")
        non_numeric_cols = df_combined.select(cs.by_dtype(pl.Utf8)).columns
        new_order = (
            sorted(non_numeric_cols)
            + ["Cell_nuclei_count", "Cell_cytoplasm_count"]
            + morphology_feature_cols
        )

        df_combined = df_combined.select(new_order)

        barcode_list = df_combined["Metadata_Barcode"].unique().to_list()
        barcode_str = ", ".join([f"'{item}'" for item in barcode_list])

        query = plate_layout_sql_query(cfg.DATABASE_SCHEMA, barcode_str)

        df_plates = pl.read_database(query, cfg.DB_URI)

        # Join data with df_plates
        df_combined = df_combined.join(
            df_plates,
            how="left",
            left_on=["Metadata_Barcode", "Metadata_Well"],
            right_on=["barcode", "well_id"],
        )
        aggregated_data = df_combined.drop_nulls(subset="batch_id")

        # Mapping of aggregation levels to their grouping columns
        grouping_columns_map = {
            "cell": [
                "CellID",
                "ImageID",
                "Metadata_AcqID",
                "Metadata_Barcode",
                "Metadata_Well",
                "Metadata_Site",
                "batch_id",
            ],
            "site": [
                "ImageID",
                "Metadata_AcqID",
                "Metadata_Barcode",
                "Metadata_Well",
                "Metadata_Site",
                "batch_id",
            ],
            "well": ["Metadata_AcqID", "Metadata_Barcode", "Metadata_Well", "batch_id"],
            "plate": ["Metadata_AcqID", "Metadata_Barcode", "batch_id"],
            "compound": ["batch_id"],
        }
        
        if use_gpu and not has_gpu():
            raise EnvironmentError("GPU is not available on this machine. Install NVIDIA System Management Interface (nvidia-smi) to enable this check.")

        # Iterate over the levels and aggregate data progressively
        for level in ["cell", "site", "well", "plate", "compound"]:
            aggregation_func = fa.aggregate_morphology_data_gpu if use_gpu else fa.aggregate_morphology_data_cpu
            aggregated_data = aggregation_func(
                df=aggregated_data,
                columns_to_aggregate=morphology_feature_cols,
                groupby_columns=grouping_columns_map[level],
                aggregation_function=aggregation_method[level],
            )
            if aggregation_level == level:
                break

        # Write the aggregated data to a parquet file
        aggregated_data.write_parquet(output_filename)
        all_dataframes.append(aggregated_data)

    progress_bar.close()

    return (
        pl.concat(all_dataframes)
        if len(all_dataframes) > 1
        else all_dataframes[0]
    )



import os
import subprocess
import polars as pl
import pandas as pd
from typing import Union, Literal
from .logger import (
    log_error,
)

def has_gpu():
    try:
        subprocess.run(["nvidia-smi"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return True
    except FileNotFoundError:
        return False
    
def get_file_extension(filename):
    """Helper function to get file extension"""
    possible_extensions = [".parquet", ".csv", ".tsv"]
    for ext in possible_extensions:
        full_filename = filename + ext
        if os.path.isfile(full_filename):
            return ext
    log_error(
        f"No file with extensions {possible_extensions} was not found for {filename}."
    )
    return None


def read_file(filename, extension):
    """Helper function to read file based on its extension"""
    if extension == ".parquet":
        df = pl.read_parquet(filename + extension)
    elif extension in [".csv", ".tsv"]:
        delimiter = "," if extension == ".csv" else "\t"
        df = pl.read_csv(filename + extension, separator=delimiter)
    else:
        return None
    # Change column type to float32 if all values are null (unless in some case it changes to str)
    for name in df.columns:
        if df[name].is_null().sum() == len(df[name]):
            df = df.with_columns(df[name].cast(pl.Float32))
    return df

def normalize_df(df: Union[pl.DataFrame, pd.DataFrame], method: Literal["zscore", "minmax"] = "zscore"):
    # Check if data is in pandas DataFrame, if so convert to polars DataFrame
    if isinstance(df, pd.DataFrame):
        df = pl.from_pandas(df)

    methods = {
        "minmax": lambda x: (x - x.min()) / (x.max() - x.min()),
        "zscore": lambda x: (x - x.mean()) / x.std(ddof=1),
    }

    df = df.select(
        [
            (
                methods[method](df[col])
                if df[col].dtype in [pl.Float32, pl.Float64, pl.Int32, pl.Int64]
                else df[col]
            ).alias(col)
            for col in df.columns
        ]
    )
    return df



from sys import displayhook
import polars as pl
import pandas as pd
from collections import defaultdict
from typing import Union, Tuple, Literal, Set, List, Dict
import re
from .utils import normalize_df


DEFAULT_QC_MODULES = {
        "FocusScore",
        "MaxIntensity",
        "MeanIntensity",
        "PowerLogLogSlope",
        "StdIntensity",
    }


def get_qc_module(qc_data: Union[pl.DataFrame, pd.DataFrame]):
    # Collect columns related to image quality
    image_quality_cols = [col for col in qc_data.columns if "ImageQuality_" in col]
    # Remove 'ImageQuality_' prefix from column names
    image_quality_module = [
        col.replace("ImageQuality_", "") for col in image_quality_cols
    ]
    return sorted({re.sub("_.*", "", measure) for measure in image_quality_module})


def get_qc_data_dict(
    qc_data: Union[pl.DataFrame, pd.DataFrame],
    module_to_keep: Set[str] = None,
    module_to_drop: Set[str] = None,
):
    # Check if data is in pandas DataFrame, if so convert to polars DataFrame
    if isinstance(qc_data, pd.DataFrame):
        qc_data = pl.from_pandas(qc_data)

    default_module_to_keep = DEFAULT_QC_MODULES

    # If both are None, use default
    if not module_to_keep and not module_to_drop:
        module_to_keep = default_module_to_keep

    # Filter and transform column names
    image_quality_cols = [
        col for col in qc_data.columns if col.startswith("ImageQuality_")
    ]
    image_quality_measures_all = {
        col.replace("ImageQuality_", "").split("_")[0] for col in image_quality_cols
    }

    # Filter out the measures
    image_quality_measures_filtered = {
        measure
        for measure in image_quality_measures_all
        if (not module_to_keep or measure in module_to_keep)
        and (not module_to_drop or measure not in module_to_drop)
    }

    # Create the DataFrame dictionary
    return {
        measure: qc_data.select(
            [col for col in image_quality_cols if f"_{measure}" in col]
        )
        for measure in image_quality_measures_filtered
    }


def get_channels(
    qc_data: Union[pl.DataFrame, pd.DataFrame], qc_module_list: List[str] = None
):
    # Check if data is in pandas DataFrame, if so convert to polars DataFrame
    if isinstance(qc_data, pd.DataFrame):
        qc_data = pl.from_pandas(qc_data)

    pattern_digit = re.compile(r"\d+$")
    pattern_digit_letter = re.compile(r"\d+[A-Z]+")
    pattern_sub_channel = re.compile(r"^.*?_.*?_")

    if not qc_module_list:
        qc_module_list = get_qc_module(qc_data)

    result_dict = {}

    for module in qc_module_list:
        channel_list = set()
        sub_channel_list = set()
        data_dict = get_qc_data_dict(qc_data, module_to_keep={module})
        for c in data_dict[module].columns:
            if module in c:
                parts = c.split("_")
                last_part = parts[-1]
                if last_part.isdigit() or pattern_digit_letter.match(last_part):
                    channel = pattern_digit.sub("", parts[-2])
                    sub_channel = pattern_sub_channel.sub("", c)
                    channel_list.add(channel)
                    sub_channel_list.add(sub_channel)
                else:
                    channel_list.add(last_part)

        result_dict[module] = {
            "channels": sorted(channel_list),
            "sub_channels": sorted(sub_channel_list),
        }

    return result_dict


def flag_outlier_images(
    qc_data: Union[pl.DataFrame, pd.DataFrame],
    module_to_keep: Set[str] = None,
    module_to_drop: Set[str] = None,
    method: Literal["SD", "IQR"] = "SD",
    IQR_normalization: bool = True,
    normalization: Literal["zscore", "minmax"] = "zscore",
    sd_step_dict: Dict[str, Tuple[float, float]] = None,
    default_sd_step: Tuple[float, float] = (-4.5, 4.5),
    quantile_limit: float = 0.25,
    multiplier: float = 1.5,
):
    # Check if data is in pandas DataFrame, if so convert to polars DataFrame
    if isinstance(qc_data, pd.DataFrame):
        qc_data = pl.from_pandas(qc_data)

    # Create the DataFrame dictionary
    data_frame_dictionary = get_qc_data_dict(qc_data, module_to_keep, module_to_drop)
    module_list = sorted(data_frame_dictionary.keys())

    if not 0 < quantile_limit <= 0.5:
        raise ValueError("quantile_limit must be between 0 and 0.5")

    if multiplier <= 0:
        raise ValueError("multiplier must be a positive value")

    if method == "SD":
        outlier_prefix = "OutlierSD_"
        all_sd_step_dict = defaultdict(lambda: default_sd_step)

        if sd_step_dict:
            for key, value in sd_step_dict.items():
                all_sd_step_dict[key] = value

        for image_quality_name in module_list:
            # Get the current dataframe from the dictionary
            current_dataframe = data_frame_dictionary[image_quality_name]

            # Scale the dataframe values
            current_dataframe_scaled = normalize_df(
                current_dataframe, method=normalization
            )

            # Get the lower and upper treshold for the current image_quality_name
            lower_threshold, upper_threshold = all_sd_step_dict[image_quality_name]

            # Create a new flag
            new_flag_scaled = f"{outlier_prefix}{image_quality_name}_{lower_threshold}_{upper_threshold}"
            outliers = [
                1 if i == True else 0
                for i in current_dataframe_scaled.apply(
                    lambda row: any(
                        (val < lower_threshold) | (val > upper_threshold) for val in row
                    )
                ).to_series()
            ]
            qc_data = qc_data.with_columns(pl.lit(outliers).alias(new_flag_scaled))

        # Identify columns starting with 'OutlierZscore_'
        outlier_flaged_columns = [
            item for item in qc_data.columns if item.startswith(outlier_prefix)
        ]
        flagged_qc_data = qc_data.with_columns(
            pl.max(pl.col(outlier_flaged_columns)).alias("outlier_flag")
        )

    elif method == "IQR":
        outlier_prefix = "OutlierIQR_"

        for image_quality_name in module_list:
            # Get the current dataframe from the dictionary
            current_dataframe = data_frame_dictionary[image_quality_name]

            if IQR_normalization:
                # Scale the dataframe values
                current_dataframe = normalize_df(
                    current_dataframe, method=normalization
                )

            # Calculate the lower and upper quantiles
            lower_quantile = current_dataframe.quantile(quantile_limit)
            upper_quantile = current_dataframe.quantile(1 - quantile_limit)

            # Define the IQR and the bounds for outliers
            IQR = upper_quantile - lower_quantile
            lower_threshold = (lower_quantile - multiplier * IQR).to_numpy().min()
            upper_threshold = (upper_quantile + multiplier * IQR).to_numpy().max()

            # Create a new flag
            new_flag_iqr = f"{outlier_prefix}{image_quality_name}_{round(lower_threshold, 3)}_{round(upper_threshold, 3)}"
            outliers = [
                1 if i == True else 0
                for i in current_dataframe.apply(
                    lambda row: any(
                        (val < lower_threshold) | (val > upper_threshold) for val in row
                    )
                ).to_series()
            ]

            qc_data = qc_data.with_columns(pl.lit(outliers).alias(new_flag_iqr))

        # Identify columns starting with 'OutlierScaled_'
        outlier_flaged_columns = [
            item for item in qc_data.columns if item.startswith(outlier_prefix)
        ]
        flagged_qc_data = qc_data.with_columns(
            pl.max(pl.col(outlier_flaged_columns)).alias("outlier_flag")
        )

    else:
        raise ValueError("Method must be either 'zscore' or 'IQR'")

    # Display the number of flagged image based on each quality module'
    outlier_flaged_columns = [
        col for col in flagged_qc_data.columns if col.startswith(outlier_prefix)
    ]
    displayhook(flagged_qc_data.select(outlier_flaged_columns + ["outlier_flag"]).sum())

    return flagged_qc_data


import logging
from .config import LOGGING_LEVEL

# Setting up logging configuration
logging.basicConfig(level=LOGGING_LEVEL, format='%(levelname)s: %(message)s')

def set_logger_level(level):
    """
    Set the logger's level based on user input.
    
    Args:
        level (str): The desired logging level, e.g., 'DEBUG', 'INFO', 'WARNING', etc.
    """
    valid_levels = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
    
    if level.upper() not in valid_levels:
        raise ValueError(f"Invalid log level: {level}. Valid levels: {', '.join(valid_levels)}")
    
    logging.getLogger().setLevel(getattr(logging, level.upper()))

def log_debug(message):
    logging.debug(message)

def log_info(message):
    logging.info(message)

def log_warning(message):
    logging.warning(message)

def log_error(message):
    logging.error(message)

def log_critical(message):
    logging.critical(message)



import polars as pl
from pharmbio.config import DB_URI, DATABASE_SCHEMA
from pharmbio.database.queries import experiment_name_sql_query


def get_projects_list(lookup: str = None):
    query = experiment_name_sql_query(
        DATABASE_SCHEMA['EXPERIMENT_NAME_COLUMN'], DATABASE_SCHEMA['EXPERIMENT_METADATA_TABLE_NAME_ON_DB']
    )
    project_list = pl.read_database(query, DB_URI).to_dict(as_series=False)[
        DATABASE_SCHEMA['EXPERIMENT_NAME_COLUMN']
    ]

    project_list = list(filter(None, project_list))
    if lookup is not None:
        lookup = lookup.lower()
        project_list = [s for s in project_list if lookup in s.lower()]
    return project_list



import os
import re
import polars as pl
from tqdm.notebook import tqdm
from typing import Union, Tuple, Literal, Set, List, Dict
from .qc import flag_outlier_images, get_qc_module, get_qc_data_dict, get_channels
from .vs import quality_module_lineplot, plate_heatmap, COLORS
from pharmbio import config


def get_projects_list(lookup: str = None):
    query = """
        SELECT project
        FROM image_analyses_per_plate
        GROUP BY project
        ORDER BY project 
        """
    project_list = pl.read_database(query, config.DB_URI).to_dict(as_series=False)["project"]
    project_list = list(filter(None, project_list))
    if lookup is not None:
        lookup = lookup.lower()
        project_list = [s for s in project_list if lookup in s.lower()]
    return project_list


def get_image_quality_info(
    name: str,
    drop_replication: Union[str, List[int]] = "Auto",
    keep_replication: Union[str, List[int]] = "None",
    filter: dict = None,
):  # sourcery skip: low-code-quality
    # Query database and store result in Polars dataframe
    query = f"""
            SELECT *
            FROM image_analyses_per_plate
            WHERE project ILIKE '%%{name}%%'
            AND meta->>'type' = 'cp-qc'
            AND analysis_date IS NOT NULL
            ORDER BY plate_barcode 
            """
    image_quality_info_df = pl.read_database(query, DB_URI)
    data_dict = (
        image_quality_info_df.select(["project", "plate_barcode"])
        .groupby("project")
        .agg(pl.col("plate_barcode"))
        .to_dicts()
    )
    unique_project_count = image_quality_info_df.unique("project").height
    if unique_project_count == 0:
        message = f"Quering the db for {name} returned nothing."
    elif unique_project_count > 1:
        message = f"Quering the db for {name} found {unique_project_count} studies: {image_quality_info_df.unique('project')['project'].to_list()}"
    else:
        message = f"Quering the db for {name} found {unique_project_count} study: {image_quality_info_df.unique('project')['project'].to_list()}"
    print(f"{message}\n{'_'*50}")
    if unique_project_count != 0:
        for i, study in enumerate(data_dict, start=1):
            print(i)
            for value in study.values():
                print("\t" + str(value))
    print("_" * 50)
    grouped_replicates = image_quality_info_df.groupby("plate_barcode")
    for plate_name, group in grouped_replicates:
        if len(group) > 1:
            print(
                f"Analysis for the plate with barcode {plate_name} is replicated {len(group)} times with analysis_id of {sorted(group['analysis_id'].to_list())}"
            )
    if image_quality_info_df.filter(pl.col("plate_barcode").is_duplicated()).is_empty():
        print("No replicated analysis has been found!")
    if drop_replication == "Auto" and keep_replication == "None":
        # keeping the highest analysis_id value of replicated rows
        image_quality_info_df = (
            image_quality_info_df.sort("analysis_id", descending=True)
            .unique("plate_barcode", keep="first")
            .sort("analysis_id")
        )
    elif isinstance(drop_replication, list):
        # drop rows by analysis_id
        image_quality_info_df = image_quality_info_df.filter(~pl.col("analysis_id").is_in(drop_replication))
    elif isinstance(keep_replication, list):
        # drop rows by analysis_id
        image_quality_info_df = image_quality_info_df.filter(pl.col("analysis_id").is_in(keep_replication))

    if filter is None:
        return image_quality_info_df
    conditions = []
    # Iterate over each key-value pair in the filter dictionary
    for key, values in filter.items():
        # Create an OR condition for each value associated with a key
        key_conditions = [pl.col(key).str.contains(val) for val in values]
        combined_key_condition = key_conditions[0]
        for condition in key_conditions[1:]:
            combined_key_condition = combined_key_condition | condition
        conditions.append(combined_key_condition)
    # Combine all conditions with AND
    final_condition = conditions[0]
    for condition in conditions[1:]:
        final_condition = final_condition & condition
    # Apply the condition to the DataFrame
    return image_quality_info_df.filter(final_condition)


def _get_file_extension(filename):
    """Helper function to get file extension"""
    possible_extensions = [".parquet", ".csv", ".tsv"]
    for ext in possible_extensions:
        full_filename = filename + ext
        if os.path.isfile(full_filename):
            return ext
    print(
        f"Warning: No file with extensions {possible_extensions} was not found for {filename}."
    )
    return None


def _read_file(filename, extension):
    """Helper function to read file based on its extension"""
    if extension == ".parquet":
        df = pl.read_parquet(filename + extension)
    elif extension in [".csv", ".tsv"]:
        delimiter = "," if extension == ".csv" else "\t"
        df = pl.read_csv(filename + extension, separator=delimiter)
    else:
        return None
    # Change column type to float32 if all values are null (unless in some case it changes to str)
    for name in df.columns:
        if df[name].is_null().sum() == len(df[name]):
            df = df.with_columns(df[name].cast(pl.Float32))
    return df


def get_image_quality_data(
    filtered_image_quality_info: pl.DataFrame, force_merging_columns: Union[bool, str] = False
):
    # Add qc_file column based on 'results' and 'plate_barcode' columns
    filtered_image_quality_info = filtered_image_quality_info.with_columns(
        (pl.col("results") + "qcRAW_images_" + pl.col("plate_barcode")).alias("qc_file")
    )
    print("\n")
    # Read and process all the files in a list, skipping files not found
    dfs = []
    for row in filtered_image_quality_info.iter_rows(named=True):
        ext = _get_file_extension(row["qc_file"])
        if ext is not None:
            df = _read_file(row["qc_file"], ext)
            df = df.with_columns(
                pl.lit(row["plate_acq_id"]).alias("Metadata_AcqID"),
                pl.lit(row["plate_barcode"]).alias("Metadata_Barcode"),
            )
            # Cast all numerical f64 columns to f32
            for name, dtype in zip(df.columns, df.dtypes):
                if dtype == pl.Float64:
                    df = df.with_columns(pl.col(name).cast(pl.Float32))
                elif dtype == pl.Int64:
                    df = df.with_columns(pl.col(name).cast(pl.Int32))
            dfs.append(df)
            print(f"Successfully imported {df.shape}: {row['qc_file']}{ext}")

    if force_merging_columns == "keep":
        concat_method = "diagonal"  # keep all columns and fill missing values with null
    elif force_merging_columns == "drop":
        concat_method = (
            "vertical"  # merge dfs horizontally, only keeps matching columns
        )
        common_columns = set(dfs[0].columns)
        for df in dfs[1:]:
            common_columns.intersection_update(df.columns)
        dfs = [df.select(sorted(common_columns)) for df in dfs]
    else:
        # Check if all dataframes have the same shape, if not print a message
        if len({df.shape[1] for df in dfs}) > 1:
            print("\nDataframes have different shapes and cannot be stacked together!")
            return None
        concat_method = "vertical"  # standard vertical concatenation

    print(f"\n{'_'*50}\nQuality control data of {len(dfs)} plates imported!\n")
    # Concatenate all the dataframes at once and return it
    return (
        pl.concat(dfs, how=concat_method)
        .with_columns(
            (
                pl.col("Metadata_AcqID").cast(pl.Utf8)
                + "_"
                + pl.col("Metadata_Well")
                + "_"
                + pl.col("Metadata_Site").cast(pl.Utf8)
            ).alias("ImageID")
        )
        .sort(["Metadata_Barcode", "Metadata_Well", "Metadata_Site", "ImageID"])
        if dfs
        else None
    )


class QC:
    def __init__(
        self,
        name: str,
        drop_replication: Union[str, List[int]] = "Auto",
        keep_replication: Union[str, List[int]] = "None",
        force_merging_columns: Union[bool, str] = False,
        filter: dict = None,
    ) -> None:
        self.image_quality_info = get_image_quality_info(name, drop_replication, keep_replication, filter)
        self.image_quality_data = get_image_quality_data(self.image_quality_info, force_merging_columns=force_merging_columns)
        self.project = sorted(self.image_quality_info["project"].unique().to_list())
        self.project_name = self.project[0] if len(self.project) == 1 else None
        self.pipeline_name = sorted(self.image_quality_info["pipeline_name"].unique().to_list())
        self.analysis_date = sorted(self.image_quality_info["analysis_date"].unique().to_list())
        self.plate_barcode = sorted(self.image_quality_info["plate_barcode"].unique().to_list())
        self.plate_acq_name = sorted(self.image_quality_info["plate_acq_name"].unique().to_list())
        self.plate_acq_id = sorted(self.image_quality_info["plate_acq_id"].unique().to_list())
        self.analysis_id = sorted(self.image_quality_info["analysis_id"].unique().to_list())
        if self.image_quality_data is not None:
            self.plate_wells = (
                self.image_quality_data.select("Metadata_Well")
                .unique()
                .sort(by="Metadata_Well")
                .to_series()
                .to_list()
            )
            self.plate_rows = sorted(list({w[0] for w in self.plate_wells}))
            self.plate_columns = sorted(list({w[1:] for w in self.plate_wells}))
            self.image_quality_module = get_qc_module(self.image_quality_data)

    def flag_outlier_images(
        self,
        module_to_keep: Set[str] = None,
        module_to_drop: Set[str] = None,
        method: Literal["SD", "IQR"] = "SD",
        IQR_normalization: bool = True,
        normalization: Literal["zscore", "minmax"] = "zscore",
        sd_step_dict: Dict[str, Tuple[float, float]] = None,
        default_sd_step: Tuple[float, float] = (-4.5, 4.5),
        quantile_limit: float = 0.25,
        multiplier: float = 1.5,
    ):
        return flag_outlier_images(
            self.image_quality_data,
            module_to_keep,
            module_to_drop,
            method,
            IQR_normalization,
            normalization,
            sd_step_dict,
            default_sd_step,
            quantile_limit,
            multiplier,
        )

    def image_quality_module_data_dict(
        self,
        module_to_keep: Set[str] = None,
        module_to_drop: Set[str] = None,
    ):
        return get_qc_data_dict(
            self.image_quality_data,
            module_to_keep,
            module_to_drop,
        )

    def channels(
        self,
        qc_module_list: List[str] = None,
    ):
        d = get_channels(self.image_quality_data, qc_module_list)
        for module, data in d.items():
            print(module)
            print("  Channels:", data["channels"])
            if data["sub_channels"] != []:
                print("  Sub-channels:", data["sub_channels"])
            print()
        return d

    def plate_heatmap(
        self,
        plate_names: List[str] = None,
        subplot_num_columns: int = 2,
        plot_size: int = 400,
        measurement: str = "Count_nuclei",
    ):
        if not plate_names:
            plate_names = self.plate_barcode
        return plate_heatmap(
            self.image_quality_data,
            plate_names,
            subplot_num_columns,
            plot_size,
            measurement,
        )

    def quality_module_lineplot(
        self,
        qc_module_to_plot: Set[str] = None,
        title: str = None,
        plot_size: int = 1400,
        normalization: bool = True,
        normalization_method: Literal["zscore", "minmax"] = "zscore",
        y_axis_range: Tuple = (-5, 5),
        colors: List[str] = COLORS,
    ):
        if not title:
            title = self.project_name
        return quality_module_lineplot(
            self.image_quality_data,
            qc_module_to_plot,
            title,
            plot_size,
            normalization,
            normalization_method,
            y_axis_range,
            colors,
        )


def get_morphology_info(
    name: str,
    filter: dict = None,
):
    query = f"""
            SELECT *
            FROM image_analyses_per_plate
            WHERE project ILIKE '%%{name}%%'
            AND meta->>'type' = 'cp-features'
            AND analysis_date IS NOT NULL
            ORDER BY plate_acq_id, analysis_id
            """
    df = pl.read_database(query, DB_URI)

    if filter is None:
        return df
    conditions = []
    # Iterate over each key-value pair in the filter dictionary
    for key, values in filter.items():
        # Create an OR condition for each value associated with a key
        key_conditions = [pl.col(key).str.contains(val) for val in values]
        combined_key_condition = key_conditions[0]
        for condition in key_conditions[1:]:
            combined_key_condition = combined_key_condition | condition
        conditions.append(combined_key_condition)
    # Combine all conditions with AND
    final_condition = conditions[0]
    for condition in conditions[1:]:
        final_condition = final_condition & condition
    # Apply the condition to the DataFrame
    return df.filter(final_condition)


def get_morphology_data(
    saving_dir: str,
    cp_info_df: pl.DataFrame,
    object_names: List[str] = None,
    aggregation_method: str = "mean",
) -> pl.DataFrame:
    
    # Create output directory if it doesn't exist
    if not os.path.exists(saving_dir):
        os.makedirs(saving_dir)

    if object_names is None:
        object_names = ["featICF_nuclei", "featICF_cells", "featICF_cytoplasm"]
        
    # Dynamically change the plate_name_prefix based on the aggregation method
    if aggregation_method == "mean":
        plate_name_prefix = 'Image_Mean'
    elif aggregation_method == "median":
        plate_name_prefix = 'Image_Median'
    else:
        raise ValueError(f"Unsupported aggregation method: {aggregation_method}")

    # Set up progress bar for feedback
    total_iterations = cp_info_df.height * len(object_names)
    progress_bar = tqdm(total=total_iterations, desc="Processing")
    
    all_dataframes = []

    for index, plate_metadata in enumerate(cp_info_df.iter_rows(named=True)):
        separator = "\n" if index else ""
        print(f"{separator}{'_'*50}", flush=True)
        print(f'Processing plate {plate_metadata["plate_acq_name"]} ({index + 1} of {cp_info_df.height}):', flush=True)
        
        output_filename = f'{saving_dir}/{plate_name_prefix}_{plate_metadata["plate_acq_name"]}.parquet'
        if os.path.exists(output_filename):
            print(f'File already exists, reading data from: {output_filename}', flush=True)
            existing_df = pl.read_parquet(output_filename)
            all_dataframes.append(existing_df)
            progress_bar.update(len(object_names))
            continue

        feature_dataframes = {}
        for feature in object_names:
            feature_file_path = f"{plate_metadata['results']}{feature}.parquet"
            df_feature = pl.read_parquet(feature_file_path)
            cleaned_feature_name = re.sub('_.*', '', re.sub('featICF_', '', feature))
            df_feature.columns = [f"{col}_{cleaned_feature_name}" for col in df_feature.columns]
            
            feature_dataframes[feature] = df_feature
            print(f'\tReading features {df_feature.shape} - {feature[8:]}: \t{feature_file_path}', flush=True)
            
            progress_bar.update(1)
        
        print('Merging the data', flush=True)    
        # Join nuclei and cell data on specified columns
        df_combined = feature_dataframes['featICF_nuclei'].join(
            feature_dataframes['featICF_cells'],
            left_on=['Metadata_Barcode_nuclei', 'Metadata_Site_nuclei', 'Metadata_Well_nuclei', 'Parent_cells_nuclei'],
            right_on=['Metadata_Barcode_cells', 'Metadata_Site_cells', 'Metadata_Well_cells', 'ObjectNumber_cells'],
            how='left', suffix='_cells'
        )

        # Further join with cytoplasm data on specified columns
        df_combined = df_combined.join(
            feature_dataframes['featICF_cytoplasm'],
            left_on=['Metadata_Barcode_nuclei', 'Metadata_Site_nuclei', 'Metadata_Well_nuclei', 'Parent_cells_nuclei'],
            right_on=['Metadata_Barcode_cytoplasm', 'Metadata_Site_cytoplasm', 'Metadata_Well_cytoplasm', 'ObjectNumber_cytoplasm'],
            how='left', suffix='_cytoplasm'
        )

        # Add plate and barcode information to the dataframe
        metadata_cols = [
            pl.lit(plate_metadata['plate_acq_id']).alias("Metadata_AcqID_nuclei"),
            pl.lit(plate_metadata['plate_barcode']).alias("Metadata_Barcode_nuclei")
        ]
        df_combined = df_combined.with_columns(metadata_cols)

        # Renaming columns for better consistency
        rename_map = {
            'Metadata_Barcode_nuclei': 'Metadata_Barcode',
            'Metadata_Well_nuclei': 'Metadata_Well',
            'Metadata_Site_nuclei': 'Metadata_Site',
            'Metadata_AcqID_nuclei': 'Metadata_AcqID'
        }
        df_combined = df_combined.rename(rename_map)

        # Create ImageID column by concatenating other columns
        image_id = (df_combined['Metadata_AcqID'] + '_' + 
                    df_combined['Metadata_Barcode'] + '_' + 
                    df_combined['Metadata_Well'] + '_' + 
                    df_combined['Metadata_Site']).alias("ImageID")
        df_combined = df_combined.with_columns([image_id])

        # Ensure data type consistency for certain columns
        cast_cols = [
            pl.col('Metadata_AcqID').cast(pl.Utf8),
            pl.col('Metadata_Site').cast(pl.Utf8),
        ]
        df_combined = df_combined.with_columns(cast_cols)

        # Group by image-related columns and aggregate the numeric columns
        numeric_columns = [name for name in df_combined.columns if df_combined[name].is_numeric()]
        grouping_cols = ['ImageID', 'Metadata_Barcode', 'Metadata_Well', 'Metadata_Site', 'Metadata_AcqID']

        if aggregation_method == 'mean':
            aggregations = [pl.col(column).mean().alias(column) for column in numeric_columns]
        elif aggregation_method == 'median':
            aggregations = [pl.col(column).median().alias(column) for column in numeric_columns]

        grouped_data = df_combined.groupby(grouping_cols).agg(*aggregations).sort('ImageID')

        grouped_data.write_parquet(output_filename)
        all_dataframes.append(grouped_data)

    progress_bar.close()

    if len(all_dataframes) > 1:
        return pl.concat(all_dataframes)
    else:
        return all_dataframes[0]
    
class CP:
    def __init__(
        self,
        name: str,
        saving_dir: str = "data",
        filter: dict = None,
        feature_names: List[str] = None,
    ) -> None:
        
        if feature_names is None:
            feature_names = ["featICF_nuclei", "featICF_cells", "featICF_cytoplasm"]



# ---------------------------------------------------------------------------- #
#                             LOGGING LEVEL SETTING                            #
# ---------------------------------------------------------------------------- #
import logging

LOGGING_LEVEL = logging.DEBUG

# ---------------------------------------------------------------------------- #
#                         DATABASE CONNECTION SETTING                          #
#  For sequrity reason the URI address is defined as an environment variable   #
#  and should be defined by user like so:                                      #
#                                                                              #
#  in Jupyter notbooke:                                                        #
#   %env DB_URI=URI_ADDRESS                                                    #
#                                                                              #
#  or using os environ module:                                                 #
#    import os                                                                 #
#    os.environ["DB_URI"] = "URI_ADDRESS"                                      #
#                                                                              #
#  or simply add it to system in bash command line:                            #
#     export DB_URI=URI_ADDRESS                                                #
# ---------------------------------------------------------------------------- #
import os

DB_URI = os.environ.get("DB_URI")

# ---------------------------------------------------------------------------- #
#                                DATABASE SCHEMA                               #
# ---------------------------------------------------------------------------- #
DATABASE_SCHEMA = {
    "EXPERIMENT_METADATA_TABLE_NAME_ON_DB":     "image_analyses_per_plate",
    "EXPERIMENT_NAME_COLUMN":                   "project",
    "EXPERIMENT_PLATE_BARCODE_COLUMN":          "plate_barcode",
    "EXPERIMENT_PLATE_ACQID_COLUMN":            "plate_acq_id",
    "EXPERIMENT_PLATE_AQNAME_COLUMN":           "plate_acq_name",
    "EXPERIMENT_ANALYSIS_ID_COLUMN":            "analysis_id",
    "EXPERIMENT_ANALYSIS_DATE_COLUMN":          "analysis_date",
    "EXPERIMENT_RESULT_DIRECTORY_COLUMN":       "results",
    "PLATE_LAYOUT_TABLE_NAME_ON_DB":            "plate_v1",
    "PLATE_LAYOUT_BARCODE_COLUMN":              "barcode",
    "PLATE_COMPOUND_NAME_COLUMN":               "batch_id",
}
    
# ----------------------- IMAGE QUALITY METADATA SCHEMA ---------------------- #
#  The dataframe that include the metadata for the image quality data and      #
#  refrence to the image quality data file and is retrived from the ImageDB.   #
# ---------------------------------------------------------------------------- #
IMAHGE_QUALITY_METADATA_TYPE = "cp-qc"

# ------------------------- IMAGE QUALITY MODULE DATA ------------------------ #
#  Raw file prefix that includes image quality data produced by cellprofiler   #
#  for each plate and retrived from csv/parquet file in the result directory.  #
# ---------------------------------------------------------------------------- #
IMAGE_QUALITY_FILE_PREFIX = "qcRAW_images_"

# ---------------------- CELL MORPHOLOGY METADATA SCHEMA --------------------- #
#  The dataframe that include the metadata for the cell morphology data and    #
#  refrence to the cell morphology data files and is retrived from the ImageDB.#
# ---------------------------------------------------------------------------- #
CELL_MORPHOLOGY_METADATA_TYPE = "cp-features"

# ------------------------ OBJECT MORPHOLOGY DATA FILE ----------------------- #
# List of raw file names that includes object features produced by cellprofiler#
#  for each plate and retrived from csv/parquet file in the result directory.  #
# ---------------------------------------------------------------------------- #
OBJECT_FILE_NAMES = ['featICF_nuclei', 'featICF_cells', 'featICF_cytoplasm']



import polars as pl
from typing import Union, List
from .. import config as cfg
from ..config import DATABASE_SCHEMA

from ..logger import (
    log_info,
    log_warning,
)

from ..database.queries import (
    experiment_metadata_sql_query,
)

from ..utils import (
    get_file_extension,
    read_file,
)


def get_image_quality_ref(
    name: str,
    drop_replication: Union[str, List[int]] = "Auto",
    keep_replication: Union[str, List[int]] = "None",
    filter: dict = None,
):  # sourcery skip: low-code-quality
    query = experiment_metadata_sql_query(
        name, DATABASE_SCHEMA, cfg.IMAHGE_QUALITY_METADATA_TYPE
    )
    image_quality_reference = pl.read_database(query, cfg.DB_URI)
    data_dict = (
        image_quality_reference.select(
            [
                DATABASE_SCHEMA["EXPERIMENT_NAME_COLUMN"],
                DATABASE_SCHEMA["EXPERIMENT_PLATE_BARCODE_COLUMN"],
            ]
        )
        .groupby(DATABASE_SCHEMA["EXPERIMENT_NAME_COLUMN"])
        .agg(pl.col(DATABASE_SCHEMA["EXPERIMENT_PLATE_BARCODE_COLUMN"]))
        .to_dicts()
    )
    unique_project_count = image_quality_reference.unique(
        DATABASE_SCHEMA["EXPERIMENT_NAME_COLUMN"]
    ).height

    if unique_project_count == 0:
        message = f"Quering the db for {name} returned nothing."
    elif unique_project_count > 1:
        message = (
            f"Quering the db for {name} found {unique_project_count} studies: "
            f"{image_quality_reference.unique(DATABASE_SCHEMA['EXPERIMENT_NAME_COLUMN'])[DATABASE_SCHEMA['EXPERIMENT_NAME_COLUMN']].to_list()}"
        )
    else:
        message = (
            f"Quering the db for {name} found {unique_project_count} study: "
            f"{image_quality_reference.unique(DATABASE_SCHEMA['EXPERIMENT_NAME_COLUMN'])[DATABASE_SCHEMA['EXPERIMENT_NAME_COLUMN']].to_list()}"
        )
    log_info(f"{message}\n{'_'*50}")

    if unique_project_count != 0:
        for i, study in enumerate(data_dict, start=1):
            log_info(i)
            for value in study.values():
                log_info("\t" + str(value))
    log_info("\n" + "_" * 50)

    grouped_replicates = image_quality_reference.groupby(
        DATABASE_SCHEMA["EXPERIMENT_PLATE_BARCODE_COLUMN"]
    )

    for plate_name, group in grouped_replicates:
        if len(group) > 1:
            log_warning(
                (
                    f"Analysis for the plate with barcode {plate_name} is replicated {len(group)} times with "
                    f"{DATABASE_SCHEMA['EXPERIMENT_ANALYSIS_ID_COLUMN']} of {sorted(group[DATABASE_SCHEMA['EXPERIMENT_ANALYSIS_ID_COLUMN']].to_list())}"
                )
            )
    if image_quality_reference.filter(
        pl.col(DATABASE_SCHEMA["EXPERIMENT_PLATE_BARCODE_COLUMN"]).is_duplicated()
    ).is_empty():
        log_info("No replicated analysis has been found!")
    if drop_replication == "Auto" and keep_replication == "None":
        # keeping the highest analysis_id value of replicated rows
        image_quality_reference = (
            image_quality_reference.sort(
                DATABASE_SCHEMA["EXPERIMENT_ANALYSIS_ID_COLUMN"], descending=True
            )
            .unique(DATABASE_SCHEMA["EXPERIMENT_PLATE_BARCODE_COLUMN"], keep="first")
            .sort(DATABASE_SCHEMA["EXPERIMENT_ANALYSIS_ID_COLUMN"])
        )
    elif isinstance(drop_replication, list):
        # drop rows by analysis_id
        image_quality_reference = image_quality_reference.filter(
            ~pl.col(DATABASE_SCHEMA["EXPERIMENT_ANALYSIS_ID_COLUMN"]).is_in(
                drop_replication
            )
        )
    elif isinstance(keep_replication, list):
        # keep rows by analysis_id
        image_quality_reference = image_quality_reference.filter(
            pl.col(DATABASE_SCHEMA["EXPERIMENT_ANALYSIS_ID_COLUMN"]).is_in(
                keep_replication
            )
        )

    if filter is None:
        return image_quality_reference
    
    conditions = []
    for key, values in filter.items():
        # Create an OR condition for each value associated with a key
        key_conditions = [pl.col(key).str.contains(val) for val in values]
        combined_key_condition = key_conditions[0]
        for condition in key_conditions[1:]:
            combined_key_condition = combined_key_condition | condition
        conditions.append(combined_key_condition)
    # Combine all conditions with AND
    final_condition = conditions[0]
    for condition in conditions[1:]:
        final_condition = final_condition & condition
    # Apply the condition to the DataFrame
    return image_quality_reference.filter(final_condition)


def get_image_quality_data(
    filtered_image_quality_info: pl.DataFrame,
    force_merging_columns: Union[bool, str] = False,
):
    # Add image_quality_data_file column based on RESULT_DIRECTORY_COLUMN and PLATE_BARCODE_COLUMN
    filtered_image_quality_info = filtered_image_quality_info.with_columns(
        (
            pl.col(DATABASE_SCHEMA["EXPERIMENT_RESULT_DIRECTORY_COLUMN"])
            + cfg.IMAGE_QUALITY_FILE_PREFIX
            + pl.col(DATABASE_SCHEMA["EXPERIMENT_PLATE_BARCODE_COLUMN"])
        ).alias("image_quality_data_file")
    )
    log_info("\n")
    # Read and process all the files in a list, skipping files not found
    dfs = []
    for row in filtered_image_quality_info.iter_rows(named=True):
        ext = get_file_extension(row["image_quality_data_file"])
        if ext is not None:
            df = read_file(row["image_quality_data_file"], ext)
            df = df.with_columns(
                pl.lit(row[DATABASE_SCHEMA["EXPERIMENT_PLATE_ACQID_COLUMN"]]).alias(
                    "Metadata_AcqID"
                ),
                pl.lit(row[DATABASE_SCHEMA["EXPERIMENT_PLATE_BARCODE_COLUMN"]]).alias(
                    "Metadata_Barcode"
                ),
            )
            # Cast all numerical f64 columns to f32
            for name, dtype in zip(df.columns, df.dtypes):
                if dtype == pl.Float64:
                    df = df.with_columns(pl.col(name).cast(pl.Float32))
                elif dtype == pl.Int64:
                    df = df.with_columns(pl.col(name).cast(pl.Int32))
            dfs.append(df)
            log_info(
                f"Successfully imported {df.shape}: {row['image_quality_data_file']}{ext}"
            )

    if force_merging_columns == "keep":
        concat_method = "diagonal"  # keep all columns and fill missing values with null
    elif force_merging_columns == "drop":
        concat_method = (
            "vertical"  # merge dfs horizontally, only keeps matching columns
        )
        common_columns = set(dfs[0].columns)
        for df in dfs[1:]:
            common_columns.intersection_update(df.columns)
        dfs = [df.select(sorted(common_columns)) for df in dfs]
    else:
        # Check if all dataframes have the same shape, if not print a message
        if len({df.shape[1] for df in dfs}) > 1:
            log_warning(
                "\nDataframes have different shapes and cannot be stacked together!"
            )
            return None
        concat_method = "vertical"  # standard vertical concatenation

    log_info(f"\n{'_'*50}\nQuality control data of {len(dfs)} plates imported!\n")
    # Concatenate all the dataframes at once and return it
    return (
        pl.concat(dfs, how=concat_method)
        .with_columns(
            (
                pl.col("Metadata_AcqID").cast(pl.Utf8)
                + "_"
                + pl.col("Metadata_Well")
                + "_"
                + pl.col("Metadata_Site").cast(pl.Utf8)
            ).alias("ImageID")
        )
        .sort(["Metadata_Barcode", "Metadata_Well", "Metadata_Site", "ImageID"])
        # reorder columns to match desired order
        .select(pl.col(["ImageID", "Metadata_AcqID", "Metadata_Barcode", "Metadata_Well", "Metadata_Site", "ImageNumber",]),
                pl.exclude(["ImageID", "Metadata_AcqID", "Metadata_Barcode", "Metadata_Well", "Metadata_Site", "ImageNumber",])
        )
        if dfs
        else None
    )



import polars as pl


def aggregate_morphology_data_cpu(
    df, columns_to_aggregate, groupby_columns, aggregation_function="mean"
):
    grouped = df.lazy().groupby(groupby_columns)
    retain_cols = [c for c in df.columns if c not in columns_to_aggregate]
    retained_metadata_df = df.lazy().select(retain_cols)

    # Aggregate only the desired columns.
    agg_exprs = [
        getattr(pl.col(col), aggregation_function)().alias(col)
        for col in columns_to_aggregate
    ]

    # Execute the aggregation.
    agg_df = grouped.agg(agg_exprs)
    agg_df = agg_df.join(retained_metadata_df, on=groupby_columns, how="left")

    return agg_df.sort(groupby_columns).collect()


def aggregate_morphology_data_gpu(
    df, columns_to_aggregate, groupby_columns, aggregation_function="mean"
):
    try:
        import importlib
        import numpy as np

        cudf = importlib.import_module("cudf")
        # Convert the Polars DataFrame to Arrow table and then to cuDF DataFrame
        # This will avoid copying the data and thus more efficient
        arrow_table = df.to_arrow()
        df = cudf.DataFrame.from_arrow(arrow_table)
        
        # Check for special case where 'mean' should map to 'nanmean'
        if aggregation_function == "mean":
            agg_func = np.nanmean
        elif aggregation_function == "median":
            agg_func = np.nanmedian
        else:
            agg_func = getattr(np, aggregation_function)

        agg_dict = {col: agg_func for col in columns_to_aggregate}
        agg_df = df.groupby(groupby_columns).agg(agg_dict).reset_index()

        agg_df = agg_df.sort_values(by=groupby_columns)

        return pl.from_arrow(agg_df.to_arrow())

    except ImportError as e:
        raise ImportError(
            "cuDF is not available. Please install it to use GPU acceleration."
        ) from e
    except Exception as e:
        raise RuntimeError(f"An unexpected error occurred: {str(e)}") from e



from setuptools import setup, find_packages

setup(
    name="pharmbio",
    version="0.1.3",
    url="https://github.com/pharmbio/pharmbio_package",
    author="Nima Chamyani",
    author_email="nima.ch@gmail.com",
    description="This is a Python package for interacting with Pharmb.io routine analysis.",
    packages=find_packages(),
    install_requires=[
        "pandas",
        "polars",
        "pyarrow",
        "sqlalchemy",
        "connectorx>=0.3.1",
        "psycopg2-binary",
        "plotly",
        "nbformat",
        "matplotlib",
        "scikit-learn",
        "tqdm"
    ],
)



def experiment_name_sql_query(experiment_name, table_name_on_db):
    """
    Executes a SQL query to retrieve experiment names from a specified table.

    Args:
        experiment_name (str): The name of the experiment column to select.
        table_name_on_db (str): The name of the table to query.

    Returns:
        str: The SQL query string.

    Example:
        ```python
        experiment_name = 'experiment_name'
        table_name_on_db = 'experiment_table'

        query = experiment_name_sql_query(experiment_name, table_name_on_db)
        print(query)
        ```
    """
    return f"""
            SELECT {experiment_name}
            FROM {table_name_on_db}
            GROUP BY {experiment_name}
            ORDER BY {experiment_name}
            """


def experiment_metadata_sql_query(name, db_schema, experiment_type):
    """
    Executes a SQL query to retrieve experiment metadata from the database.

    Args:
        name (str): The name of the experiment to search for.
        db_schema (dict): A dictionary containing the names of the database schema tables.
        experiment_type (str): The type of experiment to filter by.

    Returns:
        str: The SQL query string.

    Example:
        ```python
        name = 'my_experiment'
        db_schema = {
            'EXPERIMENT_METADATA_TABLE_NAME_ON_DB': 'experiment_table',
            'EXPERIMENT_NAME_COLUMN': 'name',
            'EXPERIMENT_ANALYSIS_DATE_COLUMN': 'analysis_date',
            'EXPERIMENT_PLATE_BARCODE_COLUMN': 'plate_barcode',
            'EXPERIMENT_PLATE_ACQID_COLUMN': 'plate_acqid',
            'EXPERIMENT_ANALYSIS_ID_COLUMN': 'analysis_id'
        }
        experiment_type = 'cp-features'

        query = experiment_metadata_sql_query(name, db_schema, experiment_type)
        print(query)
        ```
    """
    return f"""
            SELECT *
            FROM {db_schema['EXPERIMENT_METADATA_TABLE_NAME_ON_DB']}
            WHERE {db_schema['EXPERIMENT_NAME_COLUMN']} ILIKE '%%{name}%%'
            AND meta->>'type' = '{experiment_type}'
            AND {db_schema['EXPERIMENT_ANALYSIS_DATE_COLUMN']} IS NOT NULL
            ORDER BY {db_schema['EXPERIMENT_PLATE_BARCODE_COLUMN']}, {db_schema['EXPERIMENT_PLATE_ACQID_COLUMN']}, {db_schema['EXPERIMENT_ANALYSIS_ID_COLUMN']}
            """


def plate_layout_sql_query(db_schema, plate_barcode):
    """
    Executes a SQL query to retrieve plate layout information from the database.

    Args:
        db_schema (dict): A dictionary containing the names of the database schema tables.
        plate_barcode (str): The barcode of the plate to query.

    Returns:
        str: The SQL query string.

    Example:
        ```python
        db_schema = {
            'PLATE_LAYOUT_TABLE_NAME_ON_DB': 'plate_v1',
            'PLATE_LAYOUT_BARCODE_COLUMN': 'barcode',
            'PLATE_COMPOUND_NAME_COLUMN': 'batch_id'
        }
        plate_barcode = 'ABC123'

        query = plate_layout_sql_query(db_schema, plate_barcode)
        print(query)
        ```
    """
    return f"""
            SELECT *
            FROM {db_schema['PLATE_LAYOUT_TABLE_NAME_ON_DB']}
            WHERE ({db_schema['PLATE_LAYOUT_BARCODE_COLUMN']} IN ({plate_barcode}))
            AND {db_schema['PLATE_COMPOUND_NAME_COLUMN']} <> ''
            """



import polars as pl
import pandas as pd
import plotly.figure_factory as ff
import plotly.subplots as sp
import plotly.graph_objects as go
import plotly.express as px
import numpy as np
from collections import defaultdict
from typing import Union, Literal, Tuple, Set, List, Dict
from .qc import get_channels, get_qc_data_dict, DEFAULT_QC_MODULES
from .utils import normalize_df

COLORS = px.colors.qualitative.Set1


def plate_heatmap(
    df: Union[pl.DataFrame, pd.DataFrame],
    plate_names: List[str] = None,
    subplot_num_columns: int = 2,
    plot_size: int = 400,
    measurement: str = "Count_nuclei",
    plate_well_columns: Dict[str, str] = None,
):
    if plate_well_columns is None:
        plate_well_columns = {
            "plates": "Metadata_Barcode",
            "wells": "Metadata_Well",
        }
    if isinstance(df, pd.DataFrame):
        df = pl.from_pandas(df)

    wells = (
        df.select("Metadata_Well")
        .unique()
        .sort(by="Metadata_Well")
        .to_series()
        .to_list()
    )
    rows = sorted(list({w[0] for w in wells}))
    columns = sorted(list({w[1:] for w in wells}))

    if plate_names is None:
        try:
            plate_names = (
                df.select(plate_well_columns["plates"])
                .unique()
                .sort(by=plate_well_columns["plates"])
                .to_series()
                .to_list()
            )
            print(plate_names)
        except Exception:
            print("Plate names is not specified")
            plate_names = []

    # Define the font ratio and number of rows for the grid
    font_ratio = plot_size / 400
    subplot_num_rows = -(
        -len(plate_names) // subplot_num_columns
    )  # Ceiling division to get number of rows needed

    titles = [f"{measurement} for {name} " for name in plate_names]

    # Create a subplot with subplot_num_rows rows and subplot_num_columns columns
    fig = sp.make_subplots(
        rows=subplot_num_rows,
        cols=subplot_num_columns,
        subplot_titles=titles,
    )

    for index, plate in enumerate(plate_names):
        plate_data = df.filter(pl.col(plate_well_columns["plates"]) == plate)
        heatmap_data = []
        heatmap_data_annot = []
        for row in rows:
            heatmap_row = []
            heatmap_row_annot = []
            for column in columns:
                well = row + column
                count_nuclei = plate_data.filter(
                    pl.col(plate_well_columns["wells"]) == well
                )[measurement].to_numpy()

                if count_nuclei.size == 0:
                    well_nuclei_count = 0
                else:
                    well_nuclei_count = (
                        np.mean(count_nuclei).round(decimals=0).astype(int)
                    )

                heatmap_row.append(well_nuclei_count)
                heatmap_row_annot.append(f"{well}: {well_nuclei_count}")
            heatmap_data.append(heatmap_row)
            heatmap_data_annot.append(heatmap_row_annot)

        # Calculate the subplot row and column indices
        subplot_row = index // subplot_num_columns + 1
        subplot_col = index % subplot_num_columns + 1

        heatmap = ff.create_annotated_heatmap(
            heatmap_data,
            x=[str(i + 1) for i in range(24)],
            y=rows,
            annotation_text=heatmap_data,
            colorscale="OrRd",
            hovertext=heatmap_data_annot,
            hoverinfo="text",
        )

        # Add the heatmap to the subplot
        fig.add_trace(heatmap.data[0], row=subplot_row, col=subplot_col)

    # Update x and y axes properties
    for i in fig["layout"]["annotations"]:
        i["font"] = dict(size=12 * font_ratio)
    fig.update_xaxes(tickfont=dict(size=10 * font_ratio), nticks=48, side="bottom")
    fig.update_yaxes(autorange="reversed", tickfont=dict(size=10))
    # fig.update_yaxes(tickfont=dict(size=10*font_ratio))

    # Add the new lines here to adjust annotation positions
    for ann in fig.layout.annotations:
        ann.update(y=ann.y + 0.02)

    fig.update_layout(
        height=plot_size * subplot_num_rows,
        width=plot_size * 1.425 * subplot_num_columns,
    )
    fig.show()


def _lineplot(
    data_frames: pl.DataFrame,
    colors: List[str],
    title: str,
    plot_size: int = 1400,
    normalization: bool = True,
    normalization_method: Literal["zscore", "minmax"] = "zscore",
    y_axis_range: Tuple = (-5, 5),
):
    fig = sp.make_subplots(
        rows=len(data_frames),
        cols=1,
        subplot_titles=[df[0] for df in data_frames],
        x_title="Plates",
    )

    for x in range(len(data_frames)):
        _, channel_names, raw_data = data_frames[x]
        CurrentDataFrame = (
            normalize_df(raw_data, method=normalization_method)
            if normalization
            else raw_data
        )

        min_val = CurrentDataFrame.min().to_numpy().min()  # minimum of all columns
        max_val = CurrentDataFrame.max().to_numpy().max()  # maximum of all columns
        y_axis_range = y_axis_range if normalization else (min_val, max_val)

        for i, column in enumerate(CurrentDataFrame.columns):
            channel_name = channel_names[i]
            show_in_legend = x == 0

            fig.add_trace(
                go.Scatter(
                    x=[str(j) for j in range(CurrentDataFrame.height)],
                    y=CurrentDataFrame[column],
                    mode="lines",
                    line=dict(width=0.5, color=colors[i % len(colors)]),
                    showlegend=False,
                    name=""
                    if show_in_legend
                    else channel_name,  # get the legend just for the first row
                    legendgroup=channel_name,
                ),
                row=x + 1,
                col=1,
            )

        fig.update_xaxes(
            range=[0, CurrentDataFrame.height], showticklabels=False, row=x + 1, col=1
        )
        fig.update_yaxes(range=y_axis_range, row=x + 1, col=1)

        fig.add_shape(
            type="line",
            xref="x",
            yref="paper",
            x0=CurrentDataFrame.height / 2,
            y0=y_axis_range[0],
            x1=CurrentDataFrame.height / 2,
            y1=y_axis_range[1],
            line=dict(
                color="Black",
                width=1,
                dash="dashdot",
            ),
            row=x + 1,
            col=1,
        )

    # Dummy traces for the legend
    for i, channel_name in enumerate(channel_names):
        fig.add_trace(
            go.Scatter(
                x=[None],  # these traces won't appear
                y=[None],
                mode="lines",
                line=dict(
                    width=3, color=colors[i % len(colors)]
                ),  # this will be the width in the legend
                legendgroup=channel_name,
                name=channel_name,  # this will be the name in the legend
            ),
        )

    # Add main title
    fig.update_layout(
        height=1.8 * (len(data_frames) + 1) * 100,
        title_text=title,
        title_x=0.1,
        width=plot_size,
    )

    fig.show()


def quality_module_lineplot(
    df: Union[pl.DataFrame, pd.DataFrame],
    qc_module_to_plot: Set[str] = None,
    title: str = "Unnamed",
    plot_size: int = 1400,
    normalization: bool = True,
    normalization_method: Literal["zscore", "minmax"] = "zscore",
    y_axis_range: Tuple = (-5, 5),
    colors: List[str] = COLORS,
):
    if isinstance(df, pd.DataFrame):
        df = pl.from_pandas(df)

    if not qc_module_to_plot:
        qc_module_to_plot = DEFAULT_QC_MODULES

    title = f"{title} scaled" if normalization else f"{title} raw data"
    image_quality_measures = sorted(list(qc_module_to_plot))
    data_frame_dictionary = get_qc_data_dict(df, module_to_keep=qc_module_to_plot)
    channel_dict = get_channels(df, qc_module_list=image_quality_measures)

    channel_data_frames = defaultdict(list)
    subchannel_data_frames = defaultdict(list)

    for i in image_quality_measures:
        if channel_dict[i]["sub_channels"] != []:
            channel_names = tuple(sorted(channel_dict[i]["sub_channels"]))
            subchannel_data_frames[channel_names].append(
                (i, channel_dict[i]["sub_channels"], data_frame_dictionary.get(i))
            )
        else:
            channel_names = tuple(sorted(channel_dict[i]["channels"]))
            channel_data_frames[channel_names].append(
                (i, channel_dict[i]["channels"], data_frame_dictionary.get(i))
            )

    # Then, create figures for each group of data frames
    for data_frames in channel_data_frames.values():
        _lineplot(
            data_frames,
            colors=colors,
            title=title,
            plot_size=plot_size,
            normalization=normalization,
            normalization_method=normalization_method,
            y_axis_range=y_axis_range,
        )

    for data_frames in subchannel_data_frames.values():
        _lineplot(
            data_frames,
            colors=colors,
            title=title,
            plot_size=plot_size,
            normalization=normalization,
            normalization_method=normalization_method,
            y_axis_range=y_axis_range,
        )